<?php

namespace QUI\ERP\Payments\Stripe\PaymentMethods\Recurring;

use QUI;
use QUI\ERP\Order\AbstractOrder;
use QUI\ERP\Plans\Utils as ERPPlansUtils;
use QUI\ERP\Payments\Stripe\Provider;
use QUI\ERP\Payments\Stripe\AbstractBasePayment;
use QUI\ERP\Products\Handler\Products as ProductsHandler;
use Stripe\Exception\ApiErrorException;
use Stripe\Product as StripeProduct;
use Stripe\Plan as StripePlan;
use QUI\ERP\Payments\Stripe\Utils;

/**
 * Class BillingPlans
 *
 * Manages Stripe plans for Billing
 */
class BillingPlans
{
    /**
     * Create a Stripe Billing Plan
     *
     * @param AbstractOrder $Order
     * @return string - Stripe Billing Plan ID
     *
     * @throws QUI\Exception
     * @throws ApiErrorException
     */
    public static function createBillingPlanFromOrder(AbstractOrder $Order)
    {
        if ($Order->getPaymentDataEntry(AbstractBasePayment::ATTR_STRIPE_BILLING_PLAN_ID)) {
            return $Order->getPaymentDataEntry(AbstractBasePayment::ATTR_STRIPE_BILLING_PLAN_ID);
        }

        $offerId = self::getStripeBillingPlanIdByOrder($Order);

        if ($offerId !== false) {
            return $offerId;
        }

        if (!ErpPlansUtils::isPlanOrder($Order)) {
            throw new QUI\ERP\Accounting\Payments\Exception(
                'Order #'.$Order->getHash().' contains no plan products.'
            );
        }

        // Create new Offer
        $PlanProduct = false;

        /** @var QUI\ERP\Accounting\Article $Article */
        foreach ($Order->getArticles() as $Article) {
            if ($PlanProduct === false && ErpPlansUtils::isPlanArticle($Article)) {
                $PlanProduct = ProductsHandler::getProduct($Article->getId());
                break;
            }
        }

        Provider::setupApi();

        // Read name and description from PlanProduct (= Product that contains subscription plan information)
        $Locale = $Order->getCustomer()->getLocale();

        $StripeProduct = StripeProduct::create([
            'name' => $PlanProduct->getTitle($Locale),
            'type' => 'service'
        ]);

        $planDetails          = ErpPlansUtils::getPlanDetailsFromProduct($PlanProduct);
        $invoiceIntervalParts = explode('-', $planDetails['invoice_interval']);
        $interval             = $invoiceIntervalParts[1];

        $StripePlan = StripePlan::create([
            'product'  => $StripeProduct->id,
            'nickname' => $PlanProduct->getTitle($Locale),
            'interval' => $interval,
            'amount'   => Utils::getCentAmount($Order),
            'currency' => mb_strtolower($Order->getCurrency()->getCode()),
            'metadata' => [
                'quiqqerProductId' => $PlanProduct->getId()
            ]
        ]);

        // Save reference in database
        QUI::getDataBase()->insert(
            Provider::getStripeBillingPlansTable(),
            [
                'stripe_id'           => $StripePlan->id,
                'identification_hash' => self::getIdentificationHash($Order)
            ]
        );

        return $StripePlan->id;
    }

    /**
     * Delete a Stripe billing plan (local and remote)
     *
     * @param string $planId
     * @return void
     *
     * @throws ApiErrorException
     */
    public static function deleteBillingPlan($planId)
    {
        Provider::setupApi();

        $StripePlan = StripePlan::retrieve($planId);
        $StripePlan->delete();
    }

    /**
     * Retrieves list of all Stripe billing plans via API
     *
     * @param string $startId (optional) - Stripe plan ID to start from
     * @param int $pageSize (optional) - Number of plans per page [range: 1 to 100]
     * @return array
     *
     * @throws ApiErrorException
     */
    public static function getBillingPlanList($startId = null, $pageSize = 10)
    {
        Provider::setupApi();

        if ($pageSize < 1) {
            $pageSize = 1;
        } elseif ($pageSize > 100) {
            $pageSize = 100;
        }

        $searchParams = [
            'limit' => $pageSize
        ];

        if ($startId) {
            $searchParams['starting_after'] = $startId;
        }

        return StripePlan::all($searchParams)->toArray();
    }

    /**
     * Get Stripe billing plan ID based on the articles of an order
     *
     * @param AbstractOrder $Order
     * @return string|false - ID or false if no Billing Plan exists
     */
    protected static function getStripeBillingPlanIdByOrder(AbstractOrder $Order)
    {
        try {
            $result = QUI::getDataBase()->fetch([
                'select' => ['stripe_id'],
                'from'   => Provider::getStripeBillingPlansTable(),
                'where'  => [
                    'identification_hash' => self::getIdentificationHash($Order)
                ]
            ]);
        } catch (\Exception $Exception) {
            QUI\System\Log::writeException($Exception);
            return false;
        }

        if (empty($result)) {
            return false;
        }

        return $result[0]['stripe_id'];
    }

    /**
     * Get identification hash for a Billing Plan
     *
     * @param AbstractOrder $Order
     * @return string
     * @throws QUI\Exception
     */
    protected static function getIdentificationHash(AbstractOrder $Order)
    {
        $productIds = [];

        /** @var QUI\ERP\Accounting\Article $Article */
        foreach ($Order->getArticles() as $Article) {
            $productIds[] = (int)$Article->getId();
        }

        // sort IDs ASC
        sort($productIds);

        $lang         = $Order->getCustomer()->getLang();
        $totalSum     = $Order->getPriceCalculation()->getSum()->get();
        $hashedString = $lang.$totalSum.implode(',', $productIds);

        if (Provider::isSandbox()) {
            $hashedString .= '_sandbox';
        } else {
            $hashedString .= '_production';
        }

        return hash('sha256', $hashedString);
    }
}
